from http import HTTPStatus
from .util import classproperty


# pylint: disable=too-few-public-methods
class SAFRSRelationshipObject:
    """
    Relationship object, used to emulate a SAFRSBase object for the swagger for relationship targets
    so we can call the same methods on a relationship target as we do when using SAFRSBase
    """

    _s_class_name = None
    __name__ = "name"
    http_methods = {"GET", "POST", "PATCH", "DELETE"}
    swagger_models = {"instance": None, "collection": None}

    @classmethod
    def _s_get_swagger_doc(cls, http_method):
        """Create a swagger api model based on the sqlalchemy schema
        if an instance exists in the DB, the first entry is used as example
        :param http_method: HTTP method for which to generate the doc
        :return: swagger body, responses
        """
        body = {}
        responses = {}
        object_name = cls.__name__

        object_model = {}
        responses = {str(HTTPStatus.OK.value): {"description": f"{object_name} object", "schema": object_model}}

        if http_method.upper() in ("POST", "GET"):
            responses = {
                str(HTTPStatus.OK.value): {"description": HTTPStatus.OK.description},
                str(HTTPStatus.NOT_FOUND.value): {"description": HTTPStatus.NOT_FOUND.description},
            }

        return body, responses

    @classproperty
    def _s_relationships(cls):
        """
        :return: The relationship names of the target
        """
        return cls._target._s_relationships

    @classproperty
    def _s_jsonapi_attrs(cls):
        """
        :return: target JSON:API attributes
        """
        return cls._target._s_jsonapi_attrs

    @classproperty
    def _s_type(cls):
        """
        :return: JSON:API type
        """
        return cls._target._s_type

    @classproperty
    def _s_class_name(cls):
        """
        :return: name of the target class
        """
        return cls._target.__name__



# flake8: noqa: F401
#
# The code implements some seemingly awkward constructs and redundant functionality
# This is however required for backwards compatibility, we'll get rid of it eventually
#
from .safrs_init import DB, log, SAFRS, dict_merge, test_decorator, SAFRSRequest
from .errors import ValidationError, GenericError, UnAuthorizedError, NotFoundError
from .json_encoder import SAFRSJSONEncoder, SAFRSFormattedResponse
from .safrs_api import SAFRSAPI
from .base import SAFRSBase
from .jabase import JABase
from .jsonapi_attr import jsonapi_attr
from .jsonapi import jsonapi_format_response, paginate
from .api_methods import search, startswith
from .swagger_doc import jsonapi_rpc
from .__about__ import __version__, __description__

SafrsApi = SAFRSAPI

__all__ = (
    "__version__",
    "__description__",
    #
    "SAFRSAPI",
    "SafrsApi",
    # db:
    "SAFRSBase",
    "jsonapi_attr",
    "jsonapi_rpc",
    # jsonapi:
    "SAFRSJSONEncoder",
    "paginate",
    "jsonapi_format_response",
    "SAFRSFormattedResponse",
    "JABase",
    # api_methods:
    "search",
    "startswith",
    # Errors:
    "ValidationError",
    "GenericError",
    "UnAuthorizedError",
    "NotFoundError",
    # request
    "SAFRSRequest",
)



"""
JSON:API filtering strategies
"""
from .config import get_request_param
import sqlalchemy
import safrs
from .jsonapi_attr import is_jsonapi_attr
from flask import request
from sqlalchemy.orm import joinedload


def create_query(cls):
    """
    Create a query for the target collection `cls`.
    If `include=` query parameters are given, the corresponding relationships will be joined loaded if possible
    See: https://docs.sqlalchemy.org/en/13/orm/loading_relationships.html

    :param cls: class (collection) we want to query
    """
    query = cls._s_query

    if not safrs.SAFRS.OPTIMIZED_LOADING:
        return query
    included_csv = request.args.get("include", safrs.SAFRS.DEFAULT_INCLUDED)
    included_list = [inc for inc in included_csv.split(",") if inc]

    for inc in included_list:
        current_cls = cls
        options = None
        for inc_rel_name in inc.split("."):
            if inc_rel_name not in current_cls._s_relationships:
                safrs.log.error(f"Invalid relationship : {current_cls}.{inc_rel_name}")
                break
            inc_rel = getattr(current_cls, inc_rel_name)  # == current_cls._s_relationships[inc_rel_name]
            if not hasattr(inc_rel, "lazy") or inc_rel.lazy not in ["select", "joined", "subquery", "selectin"]:
                # we can't set options for lazy_load 'dynamic'/'eager'/'raise' relationships
                # not setting them on 'noload' either
                break
            options = options.joinedload(inc_rel) if options else joinedload(inc_rel)
            current_cls = inc_rel.mapper.class_
        if options:
            query = query.options(options)

    return query


@classmethod
def jsonapi_filter(cls):
    """
    https://jsonapi.org/recommendations/#filtering
    Apply the request.args filters to the object

    :return: sqla query object
    """

    # First check if a filter= URL query parameter has been used
    # the SAFRSObject should've implemented a filter method or
    # overwritten the _s_filter method to implement custom filtering
    filter_args = get_request_param("filter")
    if filter_args:
        safrs_object_filter = getattr(cls, "filter", None)
        if isinstance(cls, (list, sqlalchemy.orm.collections.InstrumentedList)):
            # not implemented
            result = cls
        elif callable(safrs_object_filter):
            # pylint: disable=not-callable
            result = safrs_object_filter(filter_args)
        else:
            result = cls._s_filter(filter_args)
        return result

    expressions = []
    filters = get_request_param("filters", {})
    if isinstance(cls, (list, sqlalchemy.orm.collections.InstrumentedList)):
        safrs.log.debug(f"Filtering not implemented for {cls}")
        return cls

    for attr_name, val in filters.items():
        if attr_name == "id":
            return cls._s_get_instance_by_id(val)
        if attr_name not in cls._s_jsonapi_attrs:
            # validation failed: this attribute can't be queried
            safrs.log.warning(f"Invalid filter {attr_name}")
            return []
        else:
            attr = cls._s_jsonapi_attrs[attr_name]
        if is_jsonapi_attr(attr):
            # to do
            safrs.log.debug(f"Filtering not implemented for {attr}")
        else:
            expressions.append((attr, val))

    result_query = create_query(cls)
    if expressions:
        _expressions = []
        for column, val in expressions:
            if hasattr(column, "in_"):
                _expressions.append(column.in_(val.split(",")))
            else:
                safrs.log.warning(f"'{cls}.{column}' is not a column ({type(column)})")
        result_query = result_query.filter(*_expressions)

    return result_query


@classmethod
def get_swagger_filters(cls):
    """
    :return: JSON:API filters swagger spec
    create the filter[] swagger doc for all jsonapi attributes + the id

    the columns may have attributes defined that are used for custom formatting:
    - description
    - filterable
    - type
    - format
    """
    attr_list = list(cls.SAFRSObject._s_jsonapi_attrs.keys()) + ["id"]

    for attr_name in attr_list:
        # (Customizable swagger specs):
        default_filter = ""
        description = f"{attr_name} attribute filter (csv)"
        swagger_type = "string"
        swagger_format = "string"
        name_format = "filter[{}]"
        required = False

        column = getattr(cls.SAFRSObject, "_s_column_dict", {}).get(attr_name, None)
        if column is not None:
            if not getattr(column, "filterable", True):
                continue
            description = getattr(column, "description", description)
            swagger_type = getattr(column, "swagger_type", swagger_type)
            swagger_format = getattr(column, "format", swagger_format)
            name_format = getattr(column, "name_format", name_format)
            required = getattr(column, "required", required)
            default_filter = getattr(column, "default_filter", default_filter)

        param = {
            "default": default_filter,
            "type": swagger_type,
            "name": name_format.format(attr_name),
            "in": "query",
            "format": swagger_format,
            "required": required,
            "description": description,
        }
        yield param

    yield {
        "default": "",
        "type": "string",
        "name": "filter",
        "in": "query",
        "format": "string",
        "required": False,
        "description": f"Custom {cls.SAFRSObject._s_class_name} filter",
    }


class FilteringStrategy:
    def __init__(self, jsonapi_filter: classmethod = jsonapi_filter, swagger_gen: classmethod = get_swagger_filters) -> None:
        self.jsonapi_filter = jsonapi_filter
        self.swagger_gen = swagger_gen



#
# jsonapi_rpc methods that can be added to the exposed classes
#
from sqlalchemy import or_
from sqlalchemy.orm.session import make_transient
import safrs
from .jsonapi import paginate, jsonapi_sort
from .json_encoder import SAFRSFormattedResponse
from .swagger_doc import jsonapi_rpc
from .errors import GenericError, SystemValidationError


@jsonapi_rpc(http_methods=["POST"])
def duplicate(self):
    """
    description: Duplicate an object - copy it and give it a new id
    """
    session = safrs.DB.session
    session.expunge(self)
    make_transient(self)
    self.id = self.id_type()
    session.add(self)
    session.commit()
    response = SAFRSFormattedResponse(self)
    return response


@classmethod
@jsonapi_rpc(http_methods=["POST"])
def lookup_re_mysql(cls, **kwargs):  # pragma: no cover
    """
    pageable: True
    description : Regex search all matching objects (works only in MySQL!!!)
    args:
        name: thom.*
    """
    result = cls
    for key, value in kwargs.items():
        column = getattr(cls, key, None)
        if not column:
            raise SystemValidationError(f'Invalid Column "{key}"')
        try:
            result = result.query.filter(column.op("regexp")(value))
        except Exception as exc:
            raise GenericError(f"Failed to execute query {exc}")

    return SAFRSFormattedResponse(result.all())


@classmethod
@jsonapi_rpc(http_methods=["POST"])
def startswith(cls, **kwargs):  # pragma: no cover
    """
    pageable: True
    summary : lookup items where specified attributes starts with the argument string
    args:
        attr_name: value
    """
    result = cls
    response = SAFRSFormattedResponse()
    try:
        instances = result.query
        links, instances, count = paginate(instances)
        data = [item for item in instances]
        meta = {}
        errors = None
        response = SAFRSFormattedResponse(data, meta, links, errors, count)
    except Exception as exc:
        raise GenericError(f"Failed to execute query {exc}")

    for key, value in kwargs.items():
        column = getattr(cls, key, None)
        if not column:
            raise SystemValidationError(f'Invalid Column "{key}"')
        try:
            instances = result.query.filter(column.like(value + "%"))
            links, instances, count = paginate(instances)
            data = [item for item in instances]
            meta = {}
            errors = None
            response = SAFRSFormattedResponse(data, meta, links, errors, count)

        except Exception as exc:
            raise GenericError(f"Failed to execute query {exc}")
    return response


@classmethod
@jsonapi_rpc(http_methods=["POST"])
def search(cls, **kwargs):  # pragma: no cover
    """
    pageable: True
    description : lookup column names
    args:
        query: val
    """
    query = kwargs.get("query", "")
    columns = [c for c in cls._s_columns if c.type.python_type in [str, int, float]]
    if ":" in query:
        column_name, value = query.split(":")
        result = cls.query.filter(or_(column.like("%" + value + "%") for column in columns if column.name == column_name))
    else:
        result = cls.query.filter(or_(column.like("%" + query + "%") for column in columns))
    instances = jsonapi_sort(result, cls)
    links, instances, count = paginate(instances)
    data = [item for item in instances]
    meta = {}
    errors = None
    response = SAFRSFormattedResponse(data, meta, links, errors, count)
    return response


@classmethod
@jsonapi_rpc(http_methods=["POST"])
def re_search(cls, **kwargs):  # pragma: no cover
    """
    pageable: True
    description : lookup column names
    args:
        query: search.*all
    """
    query = kwargs.get("query", "")
    result = cls.query.filter(or_(column.op("regexp")(query) for column in cls._s_columns))
    instances = result
    links, instances, count = paginate(instances)
    data = [item for item in instances]
    meta = {}
    errors = None
    response = SAFRSFormattedResponse(data, meta, links, errors, count)
    return response



"""
http://jsonapi.org/format/#content-negotiation-servers

Server Responsibilities
Servers MUST send all JSON API data in response documents with the header
"Content-Type: application/vnd.api+json" without any media type parameters.

Servers MUST respond with a 415 Unsupported Media Type status code if a request specifies the header
"Content-Type: application/vnd.api+json" with any media type parameters.
This should be implemented by the app, for example using @app.before_request  and @app.after_request
"""
import re
from flask import Request, abort
from werkzeug.datastructures import TypeConversionDict
import safrs
from .config import get_config
from .errors import ValidationError
from .safrs_api import HTTP_METHODS


# pylint: disable=too-many-ancestors, logging-format-interpolation
class SAFRSRequest(Request):
    """
    Parse jsonapi-related the request arguments:
    - header: Content-Type should be "application/vnd.api+json"
    - query args: page[limit], page[offset], fields
    - body: valid json
    """

    jsonapi_content_types = ["application/json", "application/vnd.api+json"]
    is_jsonapi = False  # indicates whether this is a jsonapi request
    _extensions = set()
    filters = {}
    filter = ""  # filter is the custom filter, used as an argument by _s_filter
    includes = []
    secure = True

    def __init__(self, *args, **kwargs):
        """
        constructor
        """
        super().__init__(*args, **kwargs)
        self.parse_content_type()
        self.parse_jsonapi_args()

    def parse_content_type(self):
        """
        Check if the request content type is jsonapi and any requested extensions
        """
        if not isinstance(self.content_type, str):  # pragma: no cover
            return

        content_type = self.content_type.split(";")[0]
        if content_type not in self.jsonapi_content_types:  # pragma: no cover
            return

        self.is_jsonapi = True
        self.parameter_storage_class = TypeConversionDict

        extensions = self.content_type.split(";")[1:]
        for ext in extensions:
            ext = ext.strip().split("=")
            if ext[0] == "ext" and ext[1:]:
                ext_name = ext[1]
                self._extensions.add(ext_name)

    @property
    def page_offset(self):
        """
        :return: page offset requested by the client when fetching lists

        Json:API supports multiple paging strategies.
        (https://jsonapi.org/format/#fetching-pagination)
        Here we extract the paging parameters to be used by sqla from the url query string.
        If the client uses page[number] instead of page[offset], then we transform the
        number parameter to an offset
        """
        page_offset = self.args.get("page[offset]", 0, type=int)
        if page_offset == 0 and "page[number]" in self.args and "page[size]" in self.args:
            page_size = self.args.get("page[size]", type=int)
            page_number = self.args.get("page[number]", type=int) - 1
            page_offset = page_number * page_size
        return page_offset

    def get_page_offset(self, rel_name):
        """
            get the page offset for the included relationship resource 
            :param rel_name: name of the relationship
            :return: page offset for included resources
        """
        page_offset = self.args.get(f"page[{rel_name}][offset]", 0, type=int)
        if page_offset == 0 and "page[{rel_name}][number]" in self.args and "page[{rel_name}][size]" in self.args:
            page_size = self.args.get("page[{rel_name}][size]", type=int)
            page_number = self.args.get("page[{rel_name}][number]", type=int) - 1
            page_offset = page_number * page_size
        return page_offset

    @property
    def page_limit(self):
        """
            get the page limit for the included relationship resource 
            :param rel_name: name of the relationship
            :return: page limit for included resources
        """
        page_limit = self.args.get("page[limit]", get_config("DEFAULT_PAGE_LIMIT"), type=int)
        if "page[number]" in self.args and "page[size]" in self.args:
            return self.args.get("page[size]", type=int)
        return page_limit

    def get_page_limit(self, rel_name):
        page_limit = self.args.get(f"page[{rel_name}][limit]", self.page_limit, type=int)
        if "page[{rel_name}][number]" in self.args and "page[{rel_name}][size]" in self.args:
            return self.args.get("page[{rel_name}][size]", type=int)
        return page_limit

    @property
    def is_bulk(self):
        """
        jsonapi bulk extension, http://springbot.github.io/json-api/extensions/bulk/
        """
        return "bulk" in self._extensions

    def get_jsonapi_payload(self):
        """
        :return: jsonapi request payload
        """
        if not self.is_jsonapi:  # pragma: no cover
            safrs.log.warning(f'Invalid Media Type! "{self.content_type}"')
            # raise GenericError('Unsupported Media Type', 415)
        if self.method == "OPTIONS":
            return None
        if self.method not in HTTP_METHODS:
            abort(500)
        result = self.get_json()
        if not isinstance(result, dict):
            raise ValidationError(f"Invalid JSON Payload : {result}")
        return result

    def parse_jsonapi_args(self):
        """
        parse the jsonapi request arguments:
        - page[offset]
        - page[limit]
        - filter[]
        - fields[]
        """

        self.filters = {}
        self.fields = {}

        # Parse the jsonapi filter[] and fields[] args
        for arg, val in self.args.items():
            if arg == "filter":
                self.filter = val

            filter_attr = re.search(r"filter\[(\w+)\]", arg)
            if filter_attr:
                attr_name = filter_attr.group(1)
                self.filters[attr_name] = val

            # https://jsonapi.org/format/#fetching-sparse-fieldsets
            fields_attr = re.search(r"fields\[(\w+)\]", arg)
            if fields_attr:
                field_type = fields_attr.group(1)
                self.fields[field_type] = val.split(",")

            if arg == "include":
                self.includes = val.split(",")



# Exception Handlers
#
# The application loglevel determines the level of detail dhown to the user.
# If set to debug, too much sensitive info might be shown !
#
# The exceptions will be caught in http_method_decorator and formatted, for example:
# {
#      "title": "Authorization Error: ",
#      "detail": "Authorization Error: ",
#      "code": 401
# }
#
import traceback
from flask import request
from werkzeug.exceptions import NotFound
import safrs
from sqlalchemy.exc import DontWrapMixin
from http import HTTPStatus
from .config import is_debug

HIDDEN_LOG = "(debug logging disabled)"


class JsonapiError(Exception, DontWrapMixin):
    pass


class NotFoundError(JsonapiError, NotFound):
    """
    This exception is raised when an item was not found
    """

    status_code = HTTPStatus.NOT_FOUND.value
    message = "NotFoundError "

    def __init__(self, message="", status_code=HTTPStatus.NOT_FOUND.value, api_code=None):
        """
        :param message: Message to be returned in the (json) body
        :param status_code: HTTP Status code
        :param api_code: API code
        """
        JsonapiError.__init__(self)
        self.status_code = status_code
        safrs.log.error("Not found: %s", message)
        if is_debug():
            self.message += message
        else:
            self.message += HIDDEN_LOG


class UnAuthorizedError(JsonapiError):
    """
    This exception is raised when an authorization error occured
    we use FORBIDDEN(403) instead of UNAUTHORIZED(401) (old http status code descriptions were not clear)
    """

    status_code = HTTPStatus.FORBIDDEN.value
    message = "Authorization Error: "

    def __init__(self, message="", status_code=HTTPStatus.FORBIDDEN.value, api_code=None):
        Exception.__init__(self)
        self.status_code = status_code
        safrs.log.error("UnAuthorizedError: %s", message)
        if is_debug():
            self.message += message
        else:
            self.message += HIDDEN_LOG


class GenericError(JsonapiError):
    """
    This exception is raised when an error has been detected
    """

    status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value  # 500
    message = "Generic Error: "

    def __init__(self, message, status_code=HTTPStatus.INTERNAL_SERVER_ERROR.value, api_code=None):
        Exception.__init__(self)
        self.status_code = status_code
        safrs.log.error("Generic Error: %s", message)
        if is_debug():
            safrs.log.info(f"Error in {request.url}")
            safrs.log.debug(traceback.format_exc(120))
            self.message += str(message)
        else:
            self.message += HIDDEN_LOG


class SystemValidationError(JsonapiError):  # pragma: no cover
    """
    This exception is raised when invalid input has been detected (server side input)
    """

    status_code = HTTPStatus.BAD_REQUEST.value
    message = "Validation Error: "

    def __init__(self, message="", status_code=HTTPStatus.BAD_REQUEST.value, api_code=None):
        Exception.__init__(self)
        self.status_code = status_code
        safrs.log.error("ValidationError: %s", message)
        if is_debug():
            self.message += message
        else:
            self.message += HIDDEN_LOG


class ValidationError(JsonapiError):
    """
    This exception is raised when invalid input has been detected (client side input)
    Always send back the message to the client in the response
    """

    status_code = HTTPStatus.BAD_REQUEST.value
    message = "Validation Error: "

    def __init__(self, message="", status_code=HTTPStatus.BAD_REQUEST.value, api_code=None):
        Exception.__init__(self)
        self.status_code = status_code
        safrs.log.warning("ValidationError: %s", message)
        self.message += message



# base.py: implements the SAFRSBase SQLAlchemy db Mixin and related operations
#
# pylint: disable=logging-format-interpolation,no-self-argument,no-member,line-too-long,fixme,protected-access
#
from __future__ import annotations
import inspect
import datetime
from subprocess import list2cmdline
import sqlalchemy
import json
import operator
from http import HTTPStatus
from urllib.parse import urljoin
from flask import request, url_for, has_request_context, current_app, g
from flask_sqlalchemy import Model
from sqlalchemy.orm.session import make_transient
from sqlalchemy import inspect as sqla_inspect, or_
from sqlalchemy.orm.interfaces import ONETOMANY, MANYTOONE, MANYTOMANY
from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property
from sqlalchemy.sql.schema import Column
from functools import lru_cache

# safrs dependencies:
import safrs
import safrs.jsonapi
from .errors import GenericError, NotFoundError, ValidationError, SystemValidationError
from .safrs_types import get_id_type
from .attr_parse import parse_attr
from .config import get_config
from .jsonapi_filters import jsonapi_filter
from .jsonapi_attr import is_jsonapi_attr
from .swagger_doc import get_doc
from .util import classproperty

#
# Map SQLA types to swagger2 json types
# json supports only a couple of basic data types, which makes our job pretty easy :)
# If a type isn't found in the table, "string" will be used
# (because of this we could actually remove all the "string" types as well)
#
SQLALCHEMY_SWAGGER2_TYPE = {
    "INTEGER": "integer",
    "SMALLINT": "integer",
    "NUMERIC": "number",
    "DECIMAL": "integer",
    "VARCHAR": "string",
    "TEXT": "string",
    "DATE": "string",
    "BOOLEAN": "boolean",
    "BLOB": "string",
    "BYTEA": "string",
    "BINARY": "string",
    "VARBINARY": "string",
    "FLOAT": "number",
    "REAL": "number",
    "DATETIME": "string",
    "BIGINT": "integer",
    "ENUM": "string",
    "INTERVAL": "string",
    "CHAR": "string",
    "TIMESTAMP": "string",
    "TINYINT": "integer",
    "MEDIUMINT": "integer",
    "NVARCHAR": "string",
    "YEAR": "integer",
    "SET": "string",
    "LONGBLOB": "string",
    "TINYTEXT": "string",
    "LONGTEXT": "string",
    "MEDIUMTEXT": "string",
    "UUID": "string",
    "TIME": "string",
}
# casting of swagger types to python types
SWAGGER2_TYPE_CAST = {"integer": int, "string": str, "number": float, "boolean": bool}


#
# SAFRSBase superclass
#
class SAFRSBase(Model):
    """This SQLAlchemy mixin implements json:api serialization for SAFRS SQLalchemy Persistent Objects
    Serialization itself is performed by the ``to_dict`` method
    Initialization and instantiation are quite complex because we rely on the DB schema

    The jsonapi id is generated from the primary keys of the columns

    This class is mostly used as a sqla model mixin therefore the object attributes should not
    match column names or sqla attribute names, this is why most of the methods & properties have
    (or should have, hindsight is great :/) the distinguishing `_s_` prefix
    """

    db_commit = True  # commit instances automatically, see also _s_auto_commit property below
    url_prefix = ""
    allow_client_generated_ids = False  # Indicates whether the client is allowed to create the id
    exclude_attrs = []  # list of attribute names that should not be serialized
    exclude_rels = []  # list of relationship names that should not be serialized
    supports_includes = True  # Set to False if you don't want this class to return included items
    
    # The swagger models are kept here, this lookup table will be used when the api swagger is generated
    # on startup
    swagger_models = {"instance": None, "collection": None}
    _s_expose = True  # indicates we want to expose this (see _s_check_perms)
    jsonapi_filter = jsonapi_filter  # filtering implementation

    # Cached lookup tables
    _col_attr_name_map = None
    _attr_col_name_map = None

    # Resource classes for the collections, relationships and methods
    # overriding these allows you to extend the Resource http methods: get(), post(), patch(), delete()
    _rest_api = safrs.jsonapi.SAFRSRestAPI
    _relationship_api = safrs.jsonapi.SAFRSRestRelationshipAPI
    _rpc_api = safrs.jsonapi.SAFRSJSONRPCAPI

    _s_upsert = True  # indicates we want to lookup and use existing objects
    _s_allow_add_rels = True  # allow relationships to be added in post requests

    _s_pk_delimiter = "_"

    _s_url_root = None # url prefix shown in the "links" field, if not set, request.url_root will be 

    
    included_list = None

    def __new__(cls, *args, **kwargs):
        """
        If an object with given arguments already exists, this object is instantiated
        """
        if "id" not in kwargs or not cls._s_upsert:
            return object.__new__(cls)
        # Fetch the PKs from the kwargs so we can lookup the corresponding object
        primary_keys = cls.id_type.extract_pks(kwargs)

        # Lookup the object with the PKs
        instance = None
        try:
            instance = cls._s_query.filter_by(**primary_keys).one_or_none()
        except Exception as exc:  # pragma: no cover
            safrs.log.warning(exc)

        if instance is None:
            instance = object.__new__(cls)

        return instance

    def __init__(self, *args, **kwargs):
        """
        Object initialization, called from backend or `_s_post`
        - set the named attributes and add the object to the database
        - create relationships
        :param args:
        :param kwargs: model attributes (column & relationship values)
        """
        # All SAFRSBase subclasses have a jsonapi id, passed as "id" in web requests
        # if no id is supplied, generate a new safrs id (uuid4)
        # instantiate the id with the "id_type", this will validate the id if
        # validation is implemented
        kwargs["id"] = self.id_type(kwargs.get("id", None))

        # Initialize the attribute values: these have been passed as key-value pairs in the
        # kwargs dictionary (from json in case of a web request).
        # Retrieve the values from each attribute (== class table column)
        db_args = {}
        column_names = [c.key for c in self._s_columns]
        for name, val in kwargs.items():
            if name in self._s_relationships:
                # Add the related instances
                db_args[name] = val
            elif is_jsonapi_attr(getattr(self.__class__, name, None)):
                # Set jsonapi attributes
                attr_val = self._s_parse_attr_value(name, val)
                setattr(self, name, attr_val)
            elif name in column_names:
                # Set columns
                attr_val = self._s_parse_attr_value(name, val)
                db_args[name] = attr_val
            elif name in self.__class__._s_jsonapi_attrs:
                db_args[name] = self._s_parse_attr_value(name, val)

        # db_args now contains the class attributes. Initialize the DB model with them
        # All subclasses should have the DB.Model as superclass.
        # (SQLAlchemy doesn't work when using DB.Model as SAFRSBase superclass)
        try:
            safrs.DB.Model.__init__(self, **db_args)
        except Exception as exc:  # pragma: no cover
            # OOPS .. things are going bad, this might happen using sqla automap
            safrs.log.error(f"Failed to instantiate {self}")
            safrs.log.debug(f"db args: {db_args}")
            safrs.log.exception(exc)
            safrs.DB.Model.__init__(self)

        if self._s_auto_commit:
            # Add the object to the database if specified by the class parameters
            safrs.DB.session.add(self)
            try:
                safrs.DB.session.commit()
            except sqlalchemy.exc.SQLAlchemyError as exc:  # pragma: no cover
                # Exception may arise when a DB constrained has been violated (e.g. duplicate key)
                raise GenericError(exc)

    def __setattr__(self, attr_name, attr_val):
        """
        setattr behaves differently for `jsonapi_attr` decorated attributes
        """
        attr = self.__class__.__dict__.get(attr_name, None)
        if is_jsonapi_attr(attr) and attr.fset is None:
            # jsonapi_attr.setter not implemented for attr
            return attr_val
        else:
            return super().__setattr__(attr_name, attr_val)

    def _s_parse_attr_value(self, attr_name: str, attr_val: any):
        """
        Parse the given jsonapi attribute value so it can be stored in the db
        :param attr_name: attribute name
        :param attr_val: attribute value
        :return: parsed value
        """
        # Don't allow attributes from web requests that are not specified in _s_jsonapi_attrs
        if not has_request_context():
            # we only care about parsing when working in the request context
            return attr_val

        if attr_name == "id":
            return attr_val

        attr = self.__class__._s_jsonapi_attrs.get(attr_name, None)

        if is_jsonapi_attr(attr):
            return attr_val

        # attr is a sqlalchemy.sql.schema.Column now
        if not isinstance(attr, Column):  # pragma: no cover
            raise SystemValidationError(f"Not a column: {attr}")

        return parse_attr(attr, attr_val)

    @classmethod
    def _s_get(cls, **kwargs):
        """
        This method is called when a collection is requested with a HTTP GET to the json api
        """
        return cls.jsonapi_filter()

    @classmethod
    def _s_post(cls, jsonapi_id=None, **params) -> SAFRSBase:
        """
        This method is called when a new item is created with a POST to the json api

        :param attributes: the jsonapi "data" attributes
        :return: new `cls` instance

        `_s_post` performs attribute sanitization and calls `cls.__init__`
        The attributes may contain an "id" if `cls.allow_client_generated_ids` is True
        """
        # remove attributes that are not declared in _s_jsonapi_attrs
        attributes = {attr_name: params[attr_name] for attr_name in params if attr_name in cls._s_jsonapi_attrs}

        # Remove 'id' (or other primary keys) from the attributes, unless it is allowed by the
        # SAFRSObject allow_client_generated_ids attribute
        if cls.allow_client_generated_ids:
            # this isn't required per the jsonapi spec
            # the user may have supplied the PK in one of the attributes, in which case "id" will be ignored
            attributes["id"] = jsonapi_id if jsonapi_id is not None else params.get("id", None)
        else:
            for attr_name in attributes.copy():
                if attr_name in cls.id_type.column_names:
                    safrs.log.warning(f"Client generated IDs are not allowed ('allow_client_generated_ids' not set for {cls})")
                    del attributes[attr_name]

        # Create the object instance with the specified id and json data
        # If the instance (id) already exists, it will be updated with the data
        # pylint: disable=not-callable
        instance = cls(**attributes)

        instance._add_rels(**params)

        if not instance in safrs.DB.session:
            safrs.DB.session.add(instance)
        if not instance._s_auto_commit or sqla_inspect(instance).pending:
            #
            # The item has not yet been added/commited by the SAFRSBase,
            # in that case we have to do it ourselves
            #
            safrs.DB.session.add(instance)
            try:
                safrs.DB.session.commit()
            except sqlalchemy.exc.SQLAlchemyError as exc:  # pragma: no cover
                # Exception may arise when a db constraint has been violated
                # (e.g. duplicate key)
                safrs.log.warning(str(exc))
                raise GenericError(str(exc))

        return instance

    def _s_patch(self, **attributes) -> SAFRSBase:
        """
        Update the object attributes
        :param **attributes:
        """
        for attr_name, attr_val in attributes.items():
            if attr_name not in self.__class__._s_jsonapi_attrs:
                continue
            # check if we have permission to write
            if not self._s_check_perm(attr_name, "w"):
                continue
            attr_val = self._s_parse_attr_value(attr_name, attr_val)
            setattr(self, attr_name, attr_val)

        safrs.DB.session.commit()
        # query ourself, this will also execute sqla hooks
        return self.get_instance(self.jsonapi_id)

    def _s_delete(self) -> None:
        """
        Delete the instance from the database
        """
        safrs.DB.session.delete(self)

    def _add_rels(self, **params) -> None:
        """
        Add relationship data provided in a POST, cfr. https://jsonapi.org/format/#crud-creating
        **params contains the (HTTP POST) parameters

        only works if self._s_allow_add_rels was set.
        """

        def data2inst(data):
            subclasses = self._safrs_subclasses()
            if not (isinstance(data, dict) and "id" in data and "type" in data and data["type"] in subclasses):
                raise ValidationError(f"Invalid relationship payload: {data}")
            target_class = subclasses[data["type"]]
            return target_class._s_post(data["id"], **data.get("attributes", {}), **data.get("relationships", {}))

        for rel_name, rel_val in params.items():
            rel = self._s_relationships.get(rel_name)
            if not rel:
                continue
            if not self._s_allow_add_rels:
                raise ValidationError("Cannot add relationships (_s_allow_add_rels not set)")
            if not isinstance(rel_val, dict) or not "data" in rel_val:
                raise ValidationError(f"Invalid relationship payload: {rel_val}")
            if not self.included_list:
                self.included_list = []
            self.included_list += [rel_name]
            rel_data = rel_val["data"]
            if isinstance(rel_data, list) and rel.direction in (ONETOMANY, MANYTOMANY):
                rel_inst = [data2inst(rd) for rd in rel_data]
                setattr(self, rel_name, rel_inst)
            elif isinstance(rel_data, dict) and rel.direction == MANYTOONE:
                inst = data2inst(rel_data)
                setattr(self, rel_name, inst)
            else:
                raise ValidationError("Invalid relationship payload")

    @staticmethod
    @lru_cache(maxsize=4)
    def _safrs_subclasses():
        """
        return a dict containing all SAFRSBase subclasses
        """
        subclasses = {c._s_type: c for c in SAFRSBase.__subclasses__()}
        while True:
            cont = False
            for subclass in [sc for r in subclasses.values() for sc in r.__subclasses__()]:
                if hasattr(subclass, "_s_type") and subclass._s_type not in subclasses and Model in inspect.getmro(subclass):
                    cont = True
                    subclasses[subclass._s_type] = subclass
            if not cont:
                break
        return subclasses

    @hybrid_property
    def http_methods(self) -> list[str]:
        """
        :return: list of allowed HTTP methods
        """
        return self.__class__.http_methods

    @http_methods.expression
    def http_methods(self) -> list[str]:
        """
        :return: list of allowed HTTP methods
        """
        return ["GET", "POST", "PATCH", "DELETE", "PUT", "HEAD", "OPTIONS"]

    @classproperty
    @lru_cache(maxsize=32)
    def _s_columns(cls) -> list:
        """
        :return: list of columns that are exposed by the api
        """
        if not hasattr(cls, "__mapper__"):
            return []

        result = cls.__mapper__.columns

        if has_request_context():
            # In the web context we only return the attributes that are exposable and readable
            # i.e. where the "expose" attribute is set on the db.Column instance
            # and the "r" flag is in the permissions
            result = [c for c in result if cls._s_check_perm(cls.colname_to_attrname(c.name))]
        return result

    @hybrid_property
    def _s_relationships(self) -> dict:
        """
        :return: the relationships used for jsonapi (de/)serialization
        """
        rels = {rel.key: rel for rel in self.__mapper__.relationships if self._s_check_perm(rel.key)}
        return rels

    @_s_relationships.expression
    def _s_relationships(cls):
        """
        :return: the relationships used for jsonapi (de/)serialization
        """
        rels = {rel.key: rel for rel in cls.__mapper__.relationships if cls._s_check_perm(rel.key)}
        return rels

    @classmethod
    def colname_to_attrname(cls, col_name):
        """
        Map column name to model attribute name

        We want this:
        ```
            for attr_name, attr_val in cls.__dict__.items():
                if col_name == getattr(attr_val, "name", None):
                    return attr_name
            return col_name
        ```
        To avoid executing this loop over and over, we create a lookup table when performing the first lookup
        (this is slightly faster than using lru_cache)
        """

        if cls._col_attr_name_map is None:
            # create lookup tables for attr <-> col mapping
            cls._col_attr_name_map = {}
            cls._attr_col_name_map = {}
            for attr_name, attr_val in cls.__dict__.items():
                if attr_name.startswith("__") and attr_name.endswith("__"):
                    # skip dunder attributes
                    continue
                _col_name = getattr(attr_val, "name", attr_name)
                if attr_name == "type":
                    attr_name = "Type"
                cls._col_attr_name_map[_col_name] = attr_name
                cls._attr_col_name_map[attr_name] = _col_name

        return cls._col_attr_name_map[col_name]

    @hybrid_method
    def _s_check_perm(self, property_name, permission="r"):
        """
        Check the (instance-level) column permission
        :param column_name: column name
        :param permission: permission string (read/write)
        :return: Boolean indicating whether access is allowed
        """

        return self.__class__._s_check_perm(property_name, permission)

    @_s_check_perm.expression
    @lru_cache(maxsize=256)
    def _s_check_perm(cls, property_name, permission="r"):
        """
        Check the (class-level) column permission
        :param column_name: column name
        :param permission: permission string (read/write)
        :return: Boolean indicating whether access is allowed
        """
        if property_name.startswith("_"):
            return False

        if property_name in cls.exclude_attrs:
            return False

        if is_jsonapi_attr(cls.__dict__.get(property_name, None)):  # avoid getattr here
            return True

        if not hasattr(cls, "__mapper__"):
            # Stateless objects
            return False

        for rel in cls.__mapper__.relationships:
            if not cls.supports_includes:
                continue
            if rel.key != property_name:
                continue
            if rel.key in cls.exclude_rels:
                # relationship name has been set in exclude_rels
                return False
            if not getattr(rel.mapper.class_, "_s_expose", False):
                # only SAFRSBase instances can be exposed
                return False
            if not getattr(rel, "expose", True):
                # relationship `expose` attribute has explicitly been set to False
                return False
            return True

        for column in cls.__mapper__.columns:
            # don't expose attributes starting with an underscore
            if cls.colname_to_attrname(column.name) != property_name:
                continue
            if getattr(column, "expose", True) and permission in getattr(column, "permissions", "rw"):
                return True
            return False

        raise SystemValidationError(f"Invalid property {property_name}")

    @hybrid_property
    def _s_jsonapi_attrs(self):
        """
        :return: dictionary of exposed attribute names and values

        ---
        The `fields` variable is used to implement jsonapi "Sparse Fieldsets"
        https://jsonapi.org/format/#fetching-sparse-fieldsets:
            client MAY request that an endpoint return only specific fields in the response on a per-type basis by including a fields[TYPE] parameter.
            The value of the fields parameter MUST be a comma-separated (U+002C COMMA, “,”) list that refers to the name(s) of the fields to be returned.
            If a client requests a restricted set of fields for a given resource type, an endpoint MUST NOT include additional fields in resource objects
            of that type in its response.
        Therefore we extract the required fieldnames from the request args, eg. Users/?Users[name] => [name]
        """
        fields = self.__class__._s_jsonapi_attrs.keys()
        if has_request_context():
            fields = request.fields.get(self._s_class_name, fields)

        result = {}
        ja_attr_names = [name for name in self.__class__._s_jsonapi_attrs.keys() if self._s_check_perm(name)]

        for attr in fields:
            attr_val = ""
            attr_name = attr
            if attr in ja_attr_names:
                if hasattr(self, attr):
                    attr_val = getattr(self, attr)
                else:
                    col_name = self.colname_to_attrname(attr)
                    attr_val = getattr(self, col_name)
            try:
                # use the current_app json_encoder
                if current_app:
                    result[attr_name] = json.loads(json.dumps(attr_val, cls=current_app.json_encoder))
                else:
                    result[attr_name] = attr_val
            except UnicodeDecodeError:  # pragma: no cover
                safrs.log.warning(f"UnicodeDecodeError fetching {self}.{attr}")
                result[attr] = ""
            except Exception as exc:
                safrs.log.warning(f"Failed to fetch {self}.{attr}: {exc}")

        return result

    @_s_jsonapi_attrs.expression
    @lru_cache(maxsize=32)
    def _s_jsonapi_attrs(cls):
        """
        :return: dict of jsonapi attributes
        At the moment we expect the column name to be equal to the column name
        Things will go south if this isn't the case and we should use
        the cls.__mapper__._polymorphic_properties instead
        """
        # Cache this for better performance (a bit faster than lru_cache :)
        cached_attrs = getattr(cls, "_cached_jsonapi_attrs", None)
        if cached_attrs is not None:
            return cached_attrs

        result = {}
        for column in cls._s_columns:
            attr_name = cls.colname_to_attrname(column.name)
            if not cls._s_check_perm(attr_name):
                continue
            # jsonapi schema prohibits the use of the fields 'id' and 'type' in the attributes
            # http://jsonapi.org/format/#document-resource-object-fields
            if attr_name == "type":
                # translate type to Type
                result["Type"] = column
            elif not attr_name == "id" and attr_name not in cls._s_relationships:
                result[attr_name] = column

        for attr_name, attr_val in cls.__dict__.items():
            if is_jsonapi_attr(attr_val):
                result[attr_name] = attr_val

        cls._cached_jsonapi_attrs = result
        return result

    def _s_expunge(self):
        """
        expunge an object from its session
        """
        session = sqla_inspect(self).session
        session.expunge(self)

    @classproperty
    def _s_auto_commit(self):
        """
        :return: whether the instance should be automatically commited.
        :rtype: boolen
        fka db_commit: auto_commit is a beter name, but keep db_commit for backwards compatibility
        """
        return self.db_commit

    @_s_auto_commit.setter
    def _s_auto_commit(self, value):
        """
        :param value:
        auto_commit setter
        """
        self.db_commit = value

    def _s_clone(self, **kwargs):
        """
        Clone an object: copy the parameters and create a new id
        :param *kwargs: TBD
        """
        make_transient(self)
        # pylint: disable=attribute-defined-outside-init
        self.id = self.id_type()
        for parameter in self._s_jsonapi_attrs:
            value = kwargs.get(parameter, None)
            if value is not None:
                setattr(self, parameter, value)
        safrs.DB.session.add(self)
        if self._s_auto_commit:
            safrs.DB.session.commit()
        return self

    @classmethod
    def get_instance(cls, item=None, failsafe=False):
        """
        :param item: instance id or dict { "id" : .. "type" : ..}
        :param failsafe: indicates whether we want an exception to be raised in case the id is not found
        :return: Instance or None. An error is raised if an invalid id is used
        """
        instance = None
        # pylint: disable=invalid-name,redefined-builtin
        if isinstance(item, dict):
            id = item.get("id", None)
            if id is None:
                raise ValidationError("Invalid id")
            if item.get("type") != cls._s_type:
                raise ValidationError("Invalid item type")
        else:
            id = item
        try:
            primary_keys = cls.id_type.get_pks(id)
        except AttributeError:  # pragma: no cover
            # This happens when we request a sample from a class that is not yet loaded
            # when we're creating the swagger models
            safrs.log.debug(f'AttributeError for class "{cls.__name__}"')
            return instance  # instance is None!

        if id is not None or not failsafe:
            try:
                instance = cls._s_query.filter_by(**primary_keys).first()
            except Exception as exc:  # pragma: no cover
                raise GenericError(f"get_instance : {exc}")

            if not instance and not failsafe:
                raise NotFoundError(f'Invalid "{cls.__name__}" ID "{id}"')
        return instance

    @classmethod
    def _s_get_instance_by_id(cls, jsonapi_id):
        """
        :param jsonapi_id: jsonapi_id
        :return: query obj
        """
        primary_keys = cls.id_type.get_pks(jsonapi_id)
        return cls._s_query.filter_by(**primary_keys)

    @property
    def jsonapi_id(self):
        """
        :return: json:api id
        :rtype: str

        if the table/object has a single primary key "id", it will return this id.
        In the other cases, the jsonapi "id" will be generated by the cls.id_type (typically by combining the PKs)

        The id has to be of type string according to the jsonapi json validation schema
        """
        return str(self.id_type.get_id(self))

    @classproperty
    @lru_cache(maxsize=32)
    # pylint: disable=method-hidden
    def id_type(obj):
        """
        :return: the object's id type
        """
        id_type = get_id_type(obj, delimiter=obj._s_pk_delimiter)
        # monkey patch so we don't have to look it up next time
        obj.id_type = id_type
        return id_type

    @classproperty
    def _s_query(cls_or_self):
        """
        :return: sqla query object
        """
        result = None
        _table = getattr(cls_or_self, "_table", None)
        try:
            result = safrs.DB.session.query(cls_or_self)
        except (sqlalchemy.exc.InvalidRequestError, sqlalchemy.exc.ArgumentError) as exc:
            # this may happen when exposing a stateless object, in which case
            # the warning can be ignored.
            if getattr(cls_or_self, "_s_stateless", None):
                safrs.log.warning("Invalid SQLA request")
        except Exception as exc:
            safrs.log.exception(exc)
            safrs.log.error(exc)

        if _table:
            result = safrs.DB.session.query(_table)

        return result

    query = _s_query

    def to_dict(self, *args, **kwargs):
        """
        Create a dictionary with all the instance "attributes"
        this method will be called by SAFRSJSONEncoder to serialize objects

        :return: dictionary with jsonapi attributes
        """
        return self._s_jsonapi_attrs

    @classproperty
    def _s_class_name(cls):
        """
        :return: the name of the instances
        """
        return cls.__name__

    @classproperty
    def _s_collection_name(cls):
        """
        :return: the name of the collection, this will be used to construct the endpoint
        """
        return getattr(cls, "__tablename__", cls.__name__)

    @classproperty
    def _s_type(cls):
        """
        :return: the jsonapi "type", i.e. the tablename if this is a db model, the classname otherwise
        """
        return cls.__name__

    @hybrid_method
    def _s_jsonapi_encode(self):
        """
        :return: Encoded object according to the jsonapi specification:
        `data = {
                "attributes": { ... },
                "id": "...",
                "links": { ... },
                "relationships": { ... },
                "type": "..."
                }`
        """
        # params = { self.object_id : self.id }
        # obj_url = url_for(self.get_endpoint(), **params) # Doesn't work :(, todo : why?
        obj_url = url_for(self.get_endpoint())
        if obj_url.endswith("/"):
            obj_url = obj_url[:-1]

        self_link = self._s_url
        attributes = self.to_dict()
        relationships = self._s_get_related()
        g.ja_data.add(self)
        data = dict(attributes=attributes, id=self.jsonapi_id, links={"self": self_link}, type=self._s_type, relationships=relationships)

        return data

    def _s_get_related(self):
        """
        :return: dict of relationship names -> [related instances]

        http://jsonapi.org/format/#fetching-includes

        Inclusion of Related Resources
        Multiple related resources can be requested in a comma-separated list:
        An endpoint MAY return resources related to the primary data by default.
        An endpoint MAY also support an include request parameter to allow
        the client to customize which related resources should be returned.
        In order to request resources related to other resources,
        a dot-separated path for each relationship name can be specified

        All related instances are stored in the `Included` class so we don't have to walk
        the relationships twice

        Request parameter example:
            include=friends.books_read,friends.books_written
        """
        # included_list contains a list of relationships to include
        # it may have been set previously by Included() when called recursively
        # if it's not set, parse the include= request param here
        # included_list example: ['friends.books_read', 'friends.books_written']
        included_list = getattr(self, "included_list", None)
        if included_list is None:
            # Multiple related resources can be requested in a comma-separated list
            included_csv = request.args.get("include", safrs.SAFRS.DEFAULT_INCLUDED)
            included_list = [inc for inc in included_csv.split(",") if inc]

        excluded_csv = request.args.get("exclude", "")
        excluded_list = excluded_csv.split(",")
        # In order to recursively request related resources
        # a dot-separated path for each relationship name can be specified
        included_rels = {i.split(".")[0] for i in included_list}
        relationships = dict()

        for rel_name in included_rels:
            """
            If a server is unable to identify a relationship path or does not support inclusion of resources from a path,
            it MUST respond with 400 Bad Request.
            """
            if rel_name != safrs.SAFRS.INCLUDE_ALL and rel_name not in self._s_relationships:
                raise GenericError(f"Invalid Relationship '{rel_name}'", status_code=400)

        for rel_name, relationship in self._s_relationships.items():
            """
            http://jsonapi.org/format/#document-resource-object-relationships:

            The value of the relationships key MUST be an object (a “relationships object”).
            Members of the relationships object (“relationships”) represent
            references from the resource object in which it’s defined to other resource objects.

            Relationships may be to-one or to-many.

            A “relationship object” MUST contain at least one of the following:

            - links: a links object containing at least one of the following:
                - self: a link for the relationship itself (a “relationship link”).
                This link allows the client to directly manipulate the relationship.
                - related: a related resource link
            - data: resource linkage
            - meta: a meta object that contains non-standard meta-information
                    about the relationship.
            A relationship object that represents a to-many relationship
            MAY also contain pagination links under the links member, as described below.
            SAFRS currently implements links with self
            """
            meta = {}
            rel_name = relationship.key
            data = [] if relationship.direction in (ONETOMANY, MANYTOMANY) else None
            if rel_name in excluded_list:
                # TODO: document this
                # continue
                pass
            if rel_name in included_rels or safrs.SAFRS.INCLUDE_ALL in included_list:
                # next_included_list contains the recursive relationship names
                next_included_list = [inc_item.split(".")[1:] for inc_item in included_list if inc_item.startswith(rel_name + ".")]
                if relationship.direction == MANYTOONE:
                    # manytoone relationship contains a single instance
                    rel_item = getattr(self, rel_name)
                    if rel_item:
                        # create an Included instance that will be used for serialization eventually
                        data = Included(rel_item, next_included_list)
                elif relationship.direction in (ONETOMANY, MANYTOMANY):
                    # manytoone relationship contains a list of instances
                    # Data is optional, it's also really slow for large sets!
                    data = []
                    rel_query = getattr(self, rel_name)
                    limit = request.get_page_limit(rel_name)
                    if not get_config("ENABLE_RELATIONSHIPS"):
                        meta["warning"] = "ENABLE_RELATIONSHIPS set to false in config.py"
                    elif rel_query:
                        # todo: check if lazy=dynamic
                        # In order to work with the relationship as with Query,
                        # you need to configure it with lazy='dynamic'
                        # "limit" may not be possible !
                        if getattr(rel_query, "limit", False):
                            count = rel_query.count()
                            rel_query = rel_query.limit(limit)
                            if rel_query.count() >= get_config("BIG_QUERY_THRESHOLD"):
                                warning = f'Truncated result for relationship "{rel_name}",consider paginating this request'
                                safrs.log.warning(warning)
                                meta["warning"] = warning
                            items = rel_query.all()
                        else:  # rel_query is an 'InstrumentedList'
                            items = list(rel_query)[:limit]
                            count = len(items)
                        meta["count"] = meta["total"] = count
                        meta["limit"] = limit
                        for rel_item in items:
                            data.append(Included(rel_item, next_included_list))
                else:  # pragma: no cover
                    # should never happen
                    safrs.log.error(f"Unknown relationship direction for relationship {rel_name}: {relationship.direction}")

            rel_link = urljoin(self._s_url, rel_name)
            links = dict(self=rel_link)
            rel_data = dict(links=links)

            rel_data["data"] = data
            if meta:
                rel_data["meta"] = meta
            relationships[rel_name] = rel_data

        return relationships

    def __unicode__(self):
        """"""
        name = getattr(self, "name", self.jsonapi_id)
        return name if name is not None else ""

    __str__ = __unicode__

    @classmethod
    def _s_count(cls):
        """
        returning None will cause our jsonapi to perform a count() on the result
        this can be overridden with a cached value for performance on large tables (>1G)
        """
        max_table_count =  get_config("MAX_TABLE_COUNT")
        
        try:
            count = cls.jsonapi_filter().count()
        except Exception as exc:
            # May happen for custom types, for ex. the psycopg2 extension
            safrs.log.warning(f"Can't get count for {cls} ({exc})")
            count = -1

        if count > max_table_count:
            safrs.log.warning(f"Large table count detected ({count}>{max_table_count}), performance may be impacted, consider '{cls.__name__}._s_count' override")

        return count

    #
    # Following methods are used to create the swagger2 API documentation
    #
    @classmethod
    def _s_sample_id(cls):
        """
        :return: a sample id for the API documentation
        """
        sample = None
        if cls.query is None:
            return sample
        try:
            sample = cls.query.first()
        except Exception as exc:
            safrs.log.debug(exc)
        if sample:
            try:
                sample_id = sample.jsonapi_id
                return sample_id
            except Exception:
                safrs.log.warning(f"Failed to retrieve sample id for {cls}")

        sample_id = cls.id_type.sample_id(cls)
        return str(sample_id)  # jsonapi ids must always be strings

    @classmethod
    def _s_sample_dict(cls):
        """
        :return: a sample to be used as an example "attributes" payload in the swagger example
        """
        # create a swagger example based on the jsonapi attributes (reflecting the database column schema)
        sample = {}
        for attr_name, attr in cls._s_jsonapi_attrs.items():
            if is_jsonapi_attr(attr):
                arg = getattr(attr, "default", "")
            else:
                column = attr
                arg = None
                if hasattr(column, "sample"):
                    arg = getattr(column, "sample")
                elif hasattr(column, "default") and column.default:
                    if callable(column.default.arg):
                        # We're not executing the default when it's a callable to avoid side-effects,
                        # user may add a sample attribute to the column to have it show up in the swagger
                        safrs.log.debug(f"No OAS sample implemented for column default '{column.name}.{column.default.arg}'")
                        arg = ""
                    elif isinstance(column.type, sqlalchemy.sql.sqltypes.JSON):
                        arg = column.default.arg
                    else:
                        python_type = SWAGGER2_TYPE_CAST.get(column.type, str)
                        arg = python_type(column.default.arg)
                else:
                    # No default column value speciefd => infer one by type
                    try:
                        if column.type.python_type == int:
                            arg = 0
                        if column.type.python_type == datetime.datetime:
                            arg = str(datetime.datetime.min)
                        elif column.type.python_type == datetime.date:
                            arg = str(datetime.datetime.min.date())
                        else:
                            arg = column.type.python_type()
                    except NotImplementedError:
                        # This may happen for custom columns
                        safrs.log.debug(f"Failed to get python type for column {column} (NotImplementedError)")
                        arg = None
                    except Exception as exc:
                        safrs.log.debug(f"Failed to get python type for column {column} ({exc})")
                        # use an empty string when no type is matched, otherwise we may get json encoding
                        # errors for the swagger generation
                        arg = ""

            sample[attr_name] = arg

        return sample

    @classproperty
    def _s_object_id(cls):
        """
        :return: the Flask url parameter name of the object, e.g. UserId
        :rtype: string
        """
        # pylint: disable=no-member
        return cls.__name__ + get_config("OBJECT_ID_SUFFIX")

    @classmethod
    def _s_get_jsonapi_rpc_methods(cls):
        """
        :return: a list of jsonapi_rpc methods for this class
        :rtype: list
        """
        result = []
        try:
            cls_members = inspect.getmembers(cls)
        except sqlalchemy.exc.InvalidRequestError as exc:
            # This may happen if there's no sqlalchemy superclass
            safrs.log.warning(f"Member inspection failed for {cls}: {exc}")
            return result

        for _, method in cls_members:  # [(name, method),..]
            rest_doc = get_doc(method)
            if rest_doc is not None:
                result.append(method)
        return result

    @classmethod
    def _s_get_swagger_doc(cls, http_method):
        """
        :param http_method: the http method for which to retrieve the documentation
        :return: swagger `body` and `response` dictionaries
        :rtype: tuple
        Create a swagger api model based on the sqlalchemy schema.
        """
        body = {}
        responses = {}

        if http_method.upper() in cls.http_methods:
            responses = {HTTPStatus.NOT_FOUND.value: {"description": HTTPStatus.NOT_FOUND.description}}

            if http_method in ("post"):
                responses = {HTTPStatus.CREATED.value: {"description": HTTPStatus.CREATED.description}}

        return body, responses

    @classmethod
    def get_endpoint(cls, url_prefix=None, type=None):
        """
        :param url_prefix: URL prefix used by the app
        :param type: endpoint type, e.g. "instance"
        :return: the API endpoint
        :rtype: str
        """
        if url_prefix is None:
            url_prefix = cls.url_prefix
        if type == "instance":
            INSTANCE_ENDPOINT_FMT = get_config("INSTANCE_ENDPOINT_FMT")
            endpoint = INSTANCE_ENDPOINT_FMT.format(url_prefix, cls._s_type)
        else:  # type = 'collection'
            endpoint = f"{url_prefix}api.{cls._s_type}"
        return endpoint

    @hybrid_property
    def _s_url(self, url_prefix=""):
        """
        :param url_prefix:
        :return: endpoint url of this instance
        """
        try:
            params = {self._s_object_id: self.jsonapi_id}
            instance_url = url_for(self.get_endpoint(type="instance"), **params)
            result = urljoin(self._s_url_root, instance_url)
        except RuntimeError:
            # This happens when creating the swagger doc and there is no application registered
            result = ""
        return result

    @_s_url.expression
    def _s_url(cls, url_prefix=""):
        try:
            collection_url = url_for(cls.get_endpoint())
            result = urljoin(cls._s_url_root, collection_url)
        except RuntimeError:
            # This happens when creating the swagger doc and there is no application registered
            result = ""
        return result

    @classmethod
    def _s_meta(cls):
        """
        What is returned in the "meta" part
        may be implemented by the app
        """
        return {}

    @property
    def Type(self):
        """
        jsonapi spec doesn't allow "type" as an attribute nmae, but this is a pretty common column name
        we rename type to Type so we can support it. A bit hacky but better than not supporting "type" at all
        This may cause other errors too, for ex when sorting
        :return: renamed type
        """
        safrs.log.debug(f'({self}): attribute name "type" is reserved, renamed to "Type"')
        return self.type

    @Type.setter
    def Type(self, value):
        """
        Type property setter, see comment in the type property
        """
        if not self.Type == value:
            self.Type = value
        self.type = value

    @classmethod
    def _s_filter(cls, *filter_args, **filter_kwargs):
        """
        Apply a filter to this model
        :param filter_args: A list of filters information to apply, passed as a request URL parameter.
        Each filter object has the following fields:
        - name: The name of the field you want to filter on.
        - op: The operation you want to use (all sqlalchemy operations are available). The valid values are:
            - like: Invoke SQL like (or "ilike", "match", "notilike")
            - eq: check if field is equal to something
            - ge: check if field is greater than or equal to something
            - gt: check if field is greater than to something
            - ne: check if field is not equal to something
            - is_: check if field is a value
            - is_not: check if field is not a value
            - le: check if field is less than or equal to something
            - lt: check if field is less than to something
        - val: The value that you want to compare.
        :return: sqla query object
        """
        try:
            filters = json.loads(filter_args[0])
        except json.decoder.JSONDecodeError:
            raise ValidationError("Invalid filter format (see https://github.com/thomaxxl/safrs/wiki)")

        if not isinstance(filters, list):
            filters = [filters]

        expressions = []
        query = cls._s_query

        for filt in filters:
            if not isinstance(filt, dict):
                safrs.log.warning(f"Invalid filter '{filt}'")
                continue
            attr_name = filt.get("name")
            attr_val = filt.get("val")
            if attr_name != "id" and attr_name not in cls._s_jsonapi_attrs:
                raise ValidationError(f'Invalid filter "{filt}", unknown attribute "{attr_name}"')

            op_name = filt.get("op", "").strip("_")
            attr = cls._s_jsonapi_attrs[attr_name] if attr_name != "id" else cls.id
            if op_name in ["in", "notin"]:
                op = getattr(attr, op_name + "_")
                query = query.filter(op(attr_val))
            elif op_name in ["like", "ilike", "match", "notilike"] and hasattr(attr, "like"):
                # => attr is Column or InstrumentedAttribute
                like = getattr(attr, op_name)
                expressions.append(like(attr_val))
            elif not hasattr(operator, op_name):
                raise ValidationError(f'Invalid filter "{filt}", unknown operator "{op_name}"')
            else:
                op = getattr(operator, op_name)
                expressions.append(op(attr, attr_val))

        return query.filter(or_(*expressions))


class Included:
    """
    This class is used to serialize instances that will be included in the jsonapi response
    we keep a set of instances in `flask.g.ja_included` to avoid storing duplicates
    """

    instance = None

    def __init__(self, instance, included_list):
        """
        :param instance: the instance to be included
        :param included_list: the list of relationships that should be included for `instance` (from the url query param)
        """
        self.instance = instance
        instance.included_list = [".".join(inc_rel) for inc_rel in included_list] if included_list else []
        g.ja_included.add(instance)

    @hybrid_method
    def encode(self):
        """
        jsonapi encoding of the instance in the included relationship dictionary
        """
        return {"id": str(self.instance.jsonapi_id), "type": self.instance._s_type}

    @encode.expression
    def encode(cls):
        """
        encoding of all included instances (in the included[] part of the jsonapi response)
        """
        already_included = set()
        result = []
        while True:
            instances = getattr(g, "ja_included", None)
            if not instances:
                break
            instance = instances.pop()
            if instance in already_included or instance in g.ja_data:
                continue
            included = instance._s_jsonapi_encode()
            result.append(included)

        return result



#
# # Functions for api documentation: these decorators generate the swagger schemas
# This should evolve to a more declarative version in the future with templates
#
import inspect
import datetime
import json
import flask
from http import HTTPStatus
import yaml
from sqlalchemy.orm.interfaces import ONETOMANY, MANYTOMANY, MANYTOONE
from flask_restful_swagger_2 import Schema, swagger
from safrs.errors import SystemValidationError
from safrs.config import get_config, is_debug
import safrs
from typing import Any, Callable, Dict, List, Optional, Union


REST_DOC = "__rest_doc"  # swagger doc attribute name. If this attribute is set
# this means that the function is reachable through HTTP POST
HTTP_METHODS = "__http_method"
DOC_DELIMITER = "---"  # used as delimiter between the rest_doc swagger yaml spec
# and regular documentation
PAGEABLE = "pageable"  # denotes whether an api method is pageable
FILTERABLE = "filterable"

# additional responses added when in debug mode to make swagger-check succeed
debug_responses = {
    HTTPStatus.METHOD_NOT_ALLOWED.value: {"description": HTTPStatus.METHOD_NOT_ALLOWED.description},
    HTTPStatus.BAD_REQUEST.value: {"description": HTTPStatus.BAD_REQUEST.description},
    HTTPStatus.INTERNAL_SERVER_ERROR.value: {"description": "Internal Server Error"},
}

# List to generate the swagger references / definitions unique name
# The _references will be added to the swagger in safrs
Schema._reference_count = []
Schema._references = {}


# pylint: disable=redefined-builtin,line-too-long,protected-access,logging-format-interpolation
def parse_object_doc(object: Callable) -> Dict[str, Union[str, Dict[int, Dict[str, str]], bool, Dict[str, str]]]:
    """
    Parse the yaml description from the documented methods
    """
    api_doc = {}
    obj_doc = str(inspect.getdoc(object))
    raw_doc = obj_doc.split(DOC_DELIMITER)[0]
    yaml_doc = None

    try:
        yaml_doc = yaml.safe_load(raw_doc)
    except (SyntaxError, yaml.scanner.ScannerError) as exc:
        safrs.log.error(f"Failed to parse documentation {raw_doc} ({exc})")
        yaml_doc = {"description": raw_doc}
    except Exception:
        raise SystemValidationError("Failed to parse api doc")

    if isinstance(yaml_doc, dict):
        api_doc.update(yaml_doc)

    return api_doc


def jsonapi_rpc(http_methods: Optional[List[str]] = None, valid_jsonapi: bool = True) -> Callable:
    """
    Decorator to expose functions in the REST API:
    When a method is decorated with jsonapi_rpc, this means
    it becomes available for use through HTTP POST (i.e. public)

    :param http_methods:
    :return: function
    """
    if http_methods is None:
        http_methods = ["POST"]

    def _documented_api_method(method):
        """
        :param method:
        add metadata to the method:
            REST_DOC: swagger documentation
            HTTP_METHODS: the http methods (GET/POST/..) used to call this method
        """
        USE_API_METHODS = get_config("USE_API_METHODS")
        if USE_API_METHODS:
            try:
                api_doc = parse_object_doc(method)
            except yaml.scanner.ScannerError:
                safrs.log.error("Failed to parse documentation for %s", method)
            setattr(method, REST_DOC, api_doc)
            setattr(method, HTTP_METHODS, http_methods)
            setattr(method, "valid_jsonapi", valid_jsonapi)
        return method

    return _documented_api_method


def is_public(method):
    """
    :param method: SAFRSBase method
    :return: True or False, whether the method is to be exposed
    """

    return hasattr(method, REST_DOC)


def get_doc(method: Any) -> None:
    """
    :param  method: SAFRSBase method
    :return: OAS documentation
    """

    return getattr(method, REST_DOC, None)


def get_http_methods(method):
    """
    :param method: SAFRSBase jsonapi_rpc method
    :return: a list of http methods used to call the method (e.g. POST)
    """

    return getattr(method, HTTP_METHODS, ["POST"])


def SchemaClassFactory(name, properties):
    """
    Generate a Schema class, used to describe swagger schemas
    :param name: schema class name
    :param properties: class attributes
    :return: class
    """
    # generate a unique name to be used as a reference
    idx = Schema._reference_count.count(name)
    if idx:
        if Schema._references[name].properties == properties:
            return Schema._references[name]
        name = name + str(idx)

    for s_name, schema in Schema._references.items():
        if schema.properties == properties:
            # this detect duplicate schemas
            # ignore these for now
            pass

    Schema._reference_count.append(name)

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            # here, the properties variable is the one passed to the
            # ClassFactory call
            if key not in properties:
                raise SystemValidationError(f"Argument {key} not valid for {self.__class__.__name__}")
            setattr(self, key, value)

    new_schema_cls = type(name, (Schema,), {"__init__": __init__, "properties": properties})
    new_schema_cls.description = ""
    Schema._references[name] = new_schema_cls
    return new_schema_cls


def encode_schema(obj: Dict[str, Any]) -> Dict[str, Any]:
    """
    None aka "null" is invalid in swagger schema definition
    This breaks our samples :/
    We don't add the item to the schema if it's None
    """
    result = ""

    if obj is not None:
        try:
            result = json.loads(json.dumps(obj, cls=flask.current_app.json_encoder))
        except Exception as exc:
            safrs.log.warning(f"Json encoding failed for {obj}, type {type(obj)} ({exc})")
            result = str(obj)

    return result


# pylint: disable=redefined-builtin
def schema_from_object(name, object):
    """
    :param name:
    :param object:
    :return: swagger schema object
    """

    properties = {}

    if isinstance(object, str):
        properties = {"example": "", "type": "string"}

    elif isinstance(object, int):
        properties = {"example": 0, "type": "integer"}

    elif isinstance(object, (datetime.datetime, datetime.date, datetime.time)):
        properties = {"example": str(object), "type": "string"}

    elif isinstance(object, dict):
        for k, v in object.items():
            if isinstance(v, str):
                properties[k] = {"example": v, "type": "string"}
            elif isinstance(v, int):
                properties[k] = {"example": v, "type": "integer"}
            elif isinstance(v, (dict, list)):
                if isinstance(v, dict):
                    v = encode_schema(v)
                properties[k] = {"example": v, "type": "string"}
            elif v is None:
                # swagger doesn't allow null values
                properties[k] = {"example": "", "type": "string"}
            else:  # isinstance(object, datetime.datetime):
                properties = {"example": str(k), "type": "string"}
                safrs.log.warning("Invalid schema object type %s", type(object))
    else:
        raise SystemValidationError(f"Invalid schema object type {type(object)}")

    properties = encode_schema(properties)
    schema = SchemaClassFactory(name, properties)
    return schema


def update_response_schema(responses: Dict[str, Dict[str, str]]) -> None:
    """
    Add predefined response schemas if none is available yet
    """
    http_codes = {str(status.value): status.description for status in HTTPStatus}
    for code, response in responses.items():
        if response and not response.get("schema") and int(code) >= 400:
            jsonapi_error = {"errors": [{"title": http_codes.get(code, ""), "detail": "", "code": code}]}
            err_schema = schema_from_object(f"jsonapi_error_{code}", jsonapi_error)
            responses[code]["schema"] = err_schema


def get_swagger_doc_arguments(cls, method_name, http_method):
    """
    :param cls: class containing the method to be exposed
    :param method_name: name of the method to be exposed
    :param http_method: HTTP method used to invoke
    :return: parameters, fields, description, method

    create a schema for all methods which can be called through the
    REST POST interface

    A method is called with following JSON payload:
    ```{
        "meta"   : {
                     "args" : {
                                "parameter1" : "value1" ,
                                "parameter2" : "value2" ,
                              }
                   }
    }```

    The schema is created using the values from the jsonapi_rpc decorator,
    returned by get_doc()

    We use "meta" to remain compliant with the jsonapi schema
    """

    parameters = []
    fields = {}
    # for method_name, method in inspect.getmembers(cls, predicate=inspect.ismethod):
    for name, method in inspect.getmembers(cls):
        if name == method_name:
            break
    else:
        raise SystemValidationError(f"method {method_name} not found")

    f_args = inspect.getfullargspec(method).args
    rest_doc = get_doc(method)
    description = rest_doc.get("description", "")
    if rest_doc is not None:
        method_args = rest_doc.get("args", [])  # jsonapi_rpc "POST" method arguments
        parameters = rest_doc.get("parameters", [])  # query string parameters
        if method_args and isinstance(method_args, dict):
            if http_method == "post":
                # Post arguments, these require us to build a schema
                model_name = f"{cls.__name__}_{method_name}"
                method_field = {"method": method_name, "args": method_args}
                if getattr(method, "valid_jsonapi", True):
                    payload = schema_from_object(model_name, method_field)
                    fields["meta"] = payload
                else:
                    # todo: maybe get function args from inspect
                    for k, v in method_args.items():
                        fields[k] = {"example": v}
        if rest_doc.get(PAGEABLE):
            parameters += default_paging_parameters()
        if rest_doc.get(FILTERABLE):
            for column_name, column in cls._s_column_dict.items():
                # Expose a column if it doesn't have the "expose" attribute
                # Standard SQLA columns don't have this attibute
                # but this may have been customized by a subclass
                if getattr(column, "expose", True) and getattr(column, FILTERABLE, True):
                    description = getattr(column, "description", f"{column_name} attribute filter (csv)")
                    param = {
                        "default": "",
                        "type": "string",
                        "name": f"filter[{column_name}]",
                        "in": "query",
                        "format": "string",
                        "required": False,
                        "description": description,
                    }
                    parameters += param
    elif f_args and http_method != "options":
        safrs.log.warning(f'No documentation for method "{method_name}"')
        # jsonapi_rpc method has no documentation, generate it w/ inspect
        f_defaults = inspect.getfullargspec(method).defaults or []
        if f_args and f_args[0] in ("cls", "self"):
            f_args = f_args[1:]
        args = dict(zip(f_args, f_defaults))
        model_name = f"{cls.__name__}_{method_name}"
        # model = SchemaClassFactory(model_name, [])
        # arg_field = {"schema": model, "type": "string"} # tbd?
        method_field = {"method": method_name, "args": args}
        if getattr(method, "valid_jsonapi", True):
            fields["meta"] = schema_from_object(model_name, method_field)

    for param in parameters:
        if param.get("in") is None:
            param["in"] = "query"
        if param.get("type") is None:
            param["type"] = "string"

    return parameters, fields, description, method


#
# Decorator is called when a swagger endpoint class is instantiated
# from API.expose_object eg.
#
def swagger_doc(cls, tags=None):
    """
    swagger_doc
    """

    def swagger_doc_gen(func, instance=False):
        """
        Decorator used to document SAFRSRestAPI HTTP methods exposed in the API
        """
        default_id = cls._s_sample_id()
        class_name = cls.__name__
        collection_name = cls._s_collection_name
        http_method = func.__name__.lower()
        parameters = [
            {
                "name": cls._s_object_id,
                "in": "path",
                "type": "string",
                "default": default_id,
                "required": True,
            },  # parameter id, e.g. UserId
            {
                "name": "Content-Type",  # parameter id, e.g. UserId
                "in": "header",
                "type": "string",
                "default": "application/vnd.api+json",
                "enum": ["application/vnd.api+json", "application/json"],
                "required": True,
            },
        ]
        if tags is None:
            doc_tags = [collection_name]
        else:
            doc_tags = tags

        doc = {"tags": doc_tags}

        responses = {}
        # adhere to open api
        # the model_name will hold the OAS "$ref" schema reference
        coll_model_name = f"{class_name}_coll"  # collection model name
        inst_model_name = f"{class_name}_inst"  # instance model name

        sample_dict = cls._s_sample_dict()

        # Samples with "id" are used for GET and PATCH
        sample_instance = {"attributes": sample_dict, "type": cls._s_type, "id": cls._s_sample_id()}

        if http_method == "get":
            sample_rels = {}
            for rel_name, val in cls._s_relationships.items():
                sample_rels[rel_name] = {"data": None if val.direction is MANYTOONE else [], "links": {"self": None}}
            sample_instance["relationships"] = sample_rels

        coll_sample_data = schema_from_object(coll_model_name, {"data": [sample_instance]})
        coll_sample_data.description += f"{class_name} {http_method};"

        inst_sample_data = schema_from_object(inst_model_name, {"data": sample_instance})
        inst_sample_data.description += f"{class_name} {http_method};"

        cls.swagger_models["instance"] = inst_sample_data
        cls.swagger_models["collection"] = coll_sample_data

        if http_method == "get":
            doc["summary"] = f"Retrieve a {class_name} object"
            doc["collection_summary"] = f"Retrieve a collection of {class_name} objects"
            body, responses = cls._s_get_swagger_doc(http_method)
            responses[HTTPStatus.OK.value] = {"schema": coll_sample_data, "description": HTTPStatus.OK.description}

        elif http_method == "patch":
            post_model, responses = cls._s_get_swagger_doc(http_method)
            parameters.append(
                {
                    "name": "PATCH body",
                    "in": "body",
                    "description": f"{class_name} attributes",
                    "schema": inst_sample_data,
                    "required": True,
                }
            )
            responses[HTTPStatus.OK.value] = {"schema": inst_sample_data, "description": HTTPStatus.OK.description}

        elif http_method == "post":
            _, responses = cls._s_get_swagger_doc(http_method)
            doc["summary"] = f"Create a {class_name} object"
            # Create the default POST body schema
            sample_dict = cls._s_sample_dict()
            # The POST sample doesn't contain an "id", unless cls.allow_client_generated_ids is True
            if cls.allow_client_generated_ids:
                sample_data = schema_from_object(
                    inst_model_name, {"data": {"attributes": sample_dict, "type": cls._s_type, "id": "client_generated"}}
                )
            else:
                sample_data = schema_from_object(inst_model_name, {"data": {"attributes": sample_dict, "type": cls._s_type}})

            sample_data.description += f"{class_name} {http_method};"

            parameters.append(
                {"name": "POST body", "in": "body", "description": f"{class_name} attributes", "schema": sample_data, "required": True}
            )
            responses[HTTPStatus.CREATED.value] = {"schema": inst_sample_data, "description": HTTPStatus.CREATED.description}

        elif http_method == "delete":
            _, responses = cls._s_get_swagger_doc(http_method)
        elif http_method != "options":
            # one of 'options', 'head', 'delete'
            safrs.log.debug(f'no additional documentation for "{func}" ')

        if is_debug():
            responses.update(debug_responses)

        doc["parameters"] = parameters
        doc["responses"] = responses
        doc["produces"] = ["application/vnd.api+json"]

        method_doc = parse_object_doc(func)
        safrs.dict_merge(doc, method_doc)
        apply_fstring(doc, locals())
        update_response_schema(doc["responses"])
        return swagger.doc(doc)(func)

    return swagger_doc_gen


def swagger_relationship_doc(cls, tags=None):
    """
    swagger_relationship_doc
    """

    def swagger_doc_gen(func):
        """
        Decorator used to document relationship methods exposed in the API
        """

        parent_class = cls.relationship.parent.class_
        child_class = cls.relationship.mapper.class_
        class_name = cls.__name__
        http_method = func.__name__.lower()
        #######################################################################
        # Following will only happen when exposing an exisiting DB
        #
        if not getattr(parent_class, "object_id", None):
            parent_class._s_object_id = parent_class.__name__ + "Id"
        if not getattr(child_class, "object_id", None):
            child_class._s_object_id = child_class.__name__ + "Id"
        if not getattr(parent_class, "sample_id", None):
            setattr(parent_class, "sample_id", lambda: "")
        if not getattr(child_class, "sample_id", None):
            setattr(child_class, "sample_id", lambda: "")
        if not getattr(child_class, "get_swagger_doc", None):
            setattr(child_class, "get_swagger_doc", lambda x: (None, {}))
        #
        #######################################################################

        parameters = [
            {
                "name": parent_class._s_object_id,
                "in": "path",
                "type": "string",
                "default": parent_class._s_sample_id(),
                "description": f"{parent_class.__name__} item",
                "required": True,
            },
            {
                "name": child_class._s_object_id,
                "in": "path",
                "type": "string",
                "default": child_class._s_sample_id(),
                "description": f"{class_name} item",
                "required": True,
            },
        ]

        if tags is None:
            doc_tags = [cls._s_collection_name]
        else:
            doc_tags = tags

        doc = {"tags": doc_tags}
        doc.update(parse_object_doc(func))

        responses = {}

        # Shema names (for the swagger "references")
        model_name = f"{class_name}_rel_inst"  # instance model name
        if cls.relationship.direction in (ONETOMANY, MANYTOMANY):
            model_name = f"{class_name}_rel_coll"  # collection model name

        if http_method == "get":
            _, responses = cls._s_get_swagger_doc(http_method)
        elif http_method in ("post", "patch"):
            _, responses = cls._s_get_swagger_doc(http_method)
            child_sample_id = child_class._s_sample_id()

            _, responses = child_class._s_get_swagger_doc("patch")
            data = {"type": child_class._s_type, "id": child_sample_id}

            if cls.relationship.direction in (ONETOMANY, MANYTOMANY):
                # tomany relationships only return a 204 accepted
                data = [data]
                responses.pop(HTTPStatus.OK.value, None)

            rel_post_schema = schema_from_object(model_name, {"data": data})
            rel_post_schema.description += f"{class_name} {http_method} relationship;"
            cls.swagger_models["instance"] = rel_post_schema
            cls.swagger_models["collection"] = rel_post_schema
            parameters.append(
                {
                    "name": f"{class_name} body",
                    "in": "body",
                    "description": f"{class_name} POST model",
                    "schema": rel_post_schema,
                    "required": True,
                }
            )

            if cls.relationship.direction is MANYTOONE:
                # toone relationships return 200 OK
                responses[HTTPStatus.OK.value] = {"schema": rel_post_schema, "description": HTTPStatus.OK.description}

        elif http_method == "delete":
            child_sample_id = child_class._s_sample_id()

            _, responses = child_class._s_get_swagger_doc("patch")
            data = {"type": child_class._s_type, "id": child_sample_id}

            if cls.relationship.direction in (ONETOMANY, MANYTOMANY):
                data = [data]
            rel_del_schema = schema_from_object(model_name, {"data": data})
            parameters.append(
                {
                    "name": f"{class_name} body",
                    "in": "body",
                    "description": f"{class_name} POST model",
                    "schema": rel_del_schema,
                    "required": True,
                }
            )

        elif http_method != "options":
            # one of 'options', 'head', 'patch'
            safrs.log.info(f'no documentation for "{http_method}" ')

        doc["parameters"] = parameters
        if doc.get("responses"):
            responses.update({str(val): desc for val, desc in doc["responses"].items()})

        if is_debug():
            responses.update(debug_responses)

        doc["responses"] = responses

        parent_name = parent_class.__name__  # referenced by f-string in the jsonapi.py method docstring
        child_name = child_class.__name__  # referenced by f-string
        direction = "to-many" if cls.relationship.direction in (ONETOMANY, MANYTOMANY) else "to-one"
        apply_fstring(doc, locals())
        update_response_schema(doc["responses"])

        return swagger.doc(doc)(func)

    return swagger_doc_gen


def swagger_method_doc(cls, method_name, tags=None):
    """
    Generate decorator used to document SAFRSJSONRPCAPI instances
    """

    def swagger_doc_gen(func):
        """
        decorator
        """
        method = getattr(cls, method_name, None)
        method_doc = parse_object_doc(method)
        class_name = cls.__name__
        if tags is None:
            doc_tags = [cls._s_collection_name]
        else:
            doc_tags = tags

        responses = method_doc.get("responses", {HTTPStatus.OK.value: {"description": HTTPStatus.OK.description}})
        if is_debug():
            responses.update(debug_responses)

        summary = method_doc.get("summary", f"Invoke {class_name}.{method_name}")
        description = method_doc.get("description", summary)

        doc = {"tags": doc_tags, "description": description, "summary": summary, "responses": responses}

        model_name = f"Invoke _{class_name}_{method_name}"
        param_model = SchemaClassFactory(model_name, {})
        parameters, fields, description, method = get_swagger_doc_arguments(cls, method_name, http_method=func.__name__)

        if func.__name__ == "get":
            if not parameters:
                parameters = [
                    {"name": "varargs", "in": "query", "description": f"{method_name} arguments", "required": False, "type": "string"}
                ]
        else:
            # Retrieve the swagger schemas for the jsonapi_rpc methods from the docstring
            parameters, fields, description, method = get_swagger_doc_arguments(cls, method_name, http_method=func.__name__)
            model_name = f"{func.__name__}_{cls.__name__}_{method_name}"
            param_model = SchemaClassFactory(model_name, fields)

            parameters.append({"name": model_name, "in": "body", "description": description, "schema": param_model, "required": True})

        # URL Path Parameter
        default_id = cls._s_sample_id()
        parameters.append(
            {"name": cls._s_object_id, "in": "path", "type": "string", "default": default_id, "required": True}
        )  # parameter id, e.g. UserId
        doc["parameters"] = parameters
        doc["produces"] = ["application/vnd.api+json"]

        apply_fstring(doc, locals())
        return swagger.doc(doc)(func)

    return swagger_doc_gen


def default_paging_parameters() -> List[Dict[str, Union[int, str, bool]]]:
    """
    default_paging_parameters
    """

    parameters = []
    param = {
        "default": 0,  # The 0 isn't rendered though
        "type": "integer",
        "name": "page[offset]",
        "in": "query",
        "format": "int64",
        "required": False,
        "description": "Page offset",
    }
    parameters.append(param)

    param = {
        "default": 10,
        "type": "integer",
        "name": "page[limit]",
        "in": "query",
        "format": "int64",
        "required": False,
        "description": "Max number of items",
    }
    parameters.append(param)
    return parameters


def apply_fstring(swagger_obj, vars, k=None):
    """
    Format the f-strings in the swagger object
    """
    if isinstance(swagger_obj, str):
        result = swagger_obj
        try:
            result = swagger_obj.format(**vars)
        except Exception as exc:
            safrs.log.error(f"Failed to format ({exc})")
        return result
    elif isinstance(swagger_obj, list):
        for i in swagger_obj:
            apply_fstring(i, vars)
    elif isinstance(swagger_obj, dict):
        for k, v in swagger_obj.copy().items():
            new_v = apply_fstring(v, vars)
            if isinstance(k, int):
                # used to convert integer codes (from the responses)
                del swagger_obj[k]
                k = str(k)
            swagger_obj[k] = new_v

    return swagger_obj



import logging
import os
import sys
from flask_swagger_ui import get_swaggerui_blueprint
from flask import Flask, g
from flask.json import JSONEncoder
from flask_sqlalchemy import SQLAlchemy
from .request import SAFRSRequest
from .response import SAFRSResponse
from .json_encoder import SAFRSJSONEncoder
from .jsonapi_filters import FilteringStrategy
from functools import wraps
import safrs
import flask.app
from typing import Any, Dict, Type, Union


class SAFRS:
    """This class configures the Flask application to serve SAFRSBase instances
    :param app: a Flask application.
    :param prefix: URL prefix where the swagger should be hosted. Default is '/api'
    :param LOGLEVEL: loglevel configuration variable, values from logging module (0: trace, .. 50: critical)
    """

    # Configuration settings are stored as class variables
    MAX_PAGE_LIMIT = 100000
    DEFAULT_PAGE_LIMIT = 250
    MAX_PAGE_OFFSET = 2**31
    ENABLE_RELATIONSHIPS = True
    ENABLE_METHODS = True
    LOGLEVEL = logging.WARNING
    OBJECT_ID_SUFFIX = None
    DEFAULT_INCLUDED = ""  # change to +all to include everything (slower because relationships will be fetched)
    INSTANCE_ENDPOINT_FMT = None
    INSTANCE_URL_FMT = None
    RESOURCE_URL_FMT = None
    INSTANCEMETHOD_URL_FMT = None
    CLASSMETHOD_URL_FMT = None
    RELATIONSHIP_URL_FMT = None
    ENDPOINT_FMT = None
    MAX_TABLE_COUNT = 10**7  # table counts will become really slow for large tables, inform the user about it using this
    INCLUDE_ALL = "+all"  # include= url query argument that tells us to include all related resources
    #
    config = {}
    filtering_strategy = FilteringStrategy()

    OPTIMIZED_LOADING = True

    def __init__(self, app: flask.app.Flask, *args, **kwargs) -> None:
        """
        Constructor
        """
        self.app = app
        if app is not None:
            self.init_app(app, *args, **kwargs)

    def init_app(
        self,
        app: flask.app.Flask,
        host: str = "localhost",
        port: int = 5000,
        prefix: str = "",
        app_db: None = None,
        json_encoder: Type[SAFRSJSONEncoder] = SAFRSJSONEncoder,
        swaggerui_blueprint: bool = True,
        **kwargs,
    ) -> None:
        """
        API and application initialization
        """
        if not isinstance(app, Flask):  # pragma: no cover
            raise TypeError("'app' should be Flask.")

        if app_db is None:
            app_db = app.extensions["sqlalchemy"]

        safrs.DB = self.db = app_db

        app.json_encoder = json_encoder
        app.request_class = SAFRSRequest
        app.response_class = SAFRSResponse
        app.url_map.strict_slashes = False

        if app.config.get("DEBUG", False):
            log.setLevel(logging.DEBUG)

        # Register the API blueprint
        if swaggerui_blueprint is True:
            swaggerui_blueprint = get_swaggerui_blueprint(
                prefix, f"{prefix}/swagger.json", config={"docExpansion": "none", "defaultModelsExpandDepth": -1}
            )
            app.register_blueprint(swaggerui_blueprint, url_prefix=prefix)
            swaggerui_blueprint.json_encoder = JSONEncoder

        for conf_name, conf_val in kwargs.items():
            setattr(SAFRS, conf_name, conf_val)

        for conf_name, conf_val in app.config.items():
            setattr(SAFRS, conf_name, conf_val)

        @app.before_request
        def handle_invalid_usage():
            return

        @app.before_request
        def init_ja_data():
            # ja_data holds all data[] instances that will be encoded
            # ja_included holds all included instances
            g.ja_data = set()
            g.ja_included = set()

        # pylint: disable=unused-argument,unused-variable
        @app.teardown_appcontext
        def shutdown_session(exception=None):
            """cfr. http://flask.pocoo.org/docs/0.12/patterns/sqlalchemy/"""
            self.db.session.remove()

    @staticmethod
    def init_logging(cls: int, loglevel: int = logging.WARNING) -> logging.Logger:
        """
        Specify the log format used in the webserver logs
        The webserver will catch stdout so we redirect eveything to sys.stdout
        """
        log = logging.getLogger(__name__)
        if log.level == logging.NOTSET:
            handler = logging.StreamHandler(sys.stderr)
            formatter = logging.Formatter("[%(asctime)s] %(levelname)s: %(message)s")
            handler.setFormatter(formatter)
            log.setLevel(loglevel)
            log.addHandler(handler)
        return log


def dict_merge(
    dct: Any, merge_dct: Union[Dict[str, str], Dict[str, Union[str, Dict[int, Dict[str, str]]]], Dict[int, Dict[str, str]]]
) -> None:
    """Recursive dict merge used for creating the swagger spec.
    Inspired by :meth:``dict.update()``, instead of updating only
    top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k in merge_dct:
        if k in dct and isinstance(dct[k], dict):
            dict_merge(dct[k], merge_dct[k])
        else:
            # convert to string, for ex. http return codes
            dct[str(k)] = merge_dct[k]


def test_decorator(func):  # pragma: no cover
    """Example flask-restful decorator that can be used in the "decorators" Api argument
    cfr. https://flask-restful.readthedocs.io/en/latest/api.html#id1
    """

    @wraps(func)
    def api_wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    if func.__name__.lower() == "get":
        result = api_wrapper
        return result

    return func


#
# DB and logging initialization
#
DB = SQLAlchemy()

try:
    DEBUG = os.getenv("DEBUG", logging.WARNING)
    LOGLEVEL = int(DEBUG)
except ValueError:  # pragma: no cover
    print(f'Invalid LogLevel in DEBUG Environment Variable! "{DEBUG}"')
    LOGLEVEL = logging.INFO

log = SAFRS.init_logging(LOGLEVEL)



from .base import SAFRSBase
from .util import classproperty


class JABase(SAFRSBase):
    """
    description: Stateless class example
    """

    instances = []

    _s_url = "https://safrs/.."
    _s_relationships = {}
    _s_query = None

    def __new__(cls, *args, **kwargs):
        """
        override SAFRSBase.__new__
        """
        result = object.__new__(cls)
        cls.instances.append(result)
        return result

    def __init__(self, *args, **kwargs):
        """
        Constructor
        """
        self.id = kwargs.get("id")
        # self.name = kwargs.get("name")

    @classproperty
    def s_type(cls):
        """
        Implementations should override this to provide custom types

        :return: jsonapi type string
        """
        return "JAType_" + cls.__name__

    @classmethod
    def get(cls, *args, **kwargs):
        """
        description: JA GET
        """
        return {}

    def patch(self, *args, **kwargs):
        """
        description: JA PATCH
        """
        return {}

    def delete(self):
        """
        Called for a HTTP DELETE
        """
        return None

    @property
    def jsonapi_id(self):
        return self.id

    @classmethod
    def _s_count(cls):
        """
        jsonapi response count parameter
        """
        return 0
